# -*- coding: utf-8 -*-
"""
    :copyright: Copyright 2013-2014 by Łukasz Mierzwa
    :contact: l.mierzwa@gmail.com
"""


from __future__ import unicode_literals

import os
import shutil
import tempfile
import datetime

import pytest

from upaas.storage.local import LocalStorage
from upaas.storage.exceptions import FileNotFound
from upaas.storage.utils import find_storage_handler
from upaas.config.base import ConfigurationError


@pytest.fixture(scope="module")
def storage(request):
    directory = tempfile.mkdtemp(prefix="upaas_teststorage_")
    storage = LocalStorage({'dir': directory})

    def cleanup():
        shutil.rmtree(storage.settings.dir)
    request.addfinalizer(cleanup)

    return storage


def test_find_storage():
    assert find_storage_handler(
        'upaas.storage.local.LocalStorage', settings={'dir': '/'}) is not None


def test_find_storage_invalid():
    with pytest.raises(ConfigurationError):
        find_storage_handler('upaas.storage.local.LocalStorage')


def test_valid_settings(storage):
    assert storage.settings.dir is not None


def test_invalid_settings():
    with pytest.raises(ConfigurationError):
        LocalStorage({})


def test_config_dir_exists(storage):
    assert os.path.isdir(storage.settings.dir)


def test_config_dir_missing():
    with pytest.raises(ConfigurationError):
        LocalStorage({"dir": "/non-existing-dir"})


def test_file_exists(storage):
    file_name = "file_exists"
    open(os.path.join(storage.settings.dir, file_name), "w").close()
    assert storage.exists(file_name) is True


def test_dir_exists(storage):
    dir_name = "dir_exists"
    os.mkdir(os.path.join(storage.settings.dir, dir_name))
    assert storage.exists(dir_name) is True


def test_not_exists(storage):
    assert storage.exists("missing_file") is False


def test_get_not_exists(storage):
    with pytest.raises(FileNotFound):
        storage.get("missing file", "output")


def test_put_and_get(storage, empty_dir, empty_file):
    storage.put(empty_file, "abc")
    assert os.path.isfile(os.path.join(storage.settings.dir, "abc"))

    local_path = os.path.join(empty_dir, "xyz")
    storage.get("abc", local_path)
    assert os.path.exists(local_path)


def test_delete_not_exists(storage):
    with pytest.raises(FileNotFound):
        storage.delete("missing file")


def test_put_and_delete(storage, empty_file):
    storage.put(empty_file, "delete.me")
    assert os.path.isfile(os.path.join(storage.settings.dir, "delete.me"))

    storage.delete("delete.me")
    assert os.path.isfile(
        os.path.join(storage.settings.dir, "delete.me")) is False


def test_size_empty(storage, empty_file):
    storage.put(empty_file, "empty.file")
    assert storage.size("empty.file") == 0


def test_size_not_empty(storage, empty_file):
    with open(empty_file, 'w') as f:
        f.write('123456789')

    storage.put(empty_file, "not-empty.file")
    assert storage.size("not-empty.file") == 9


def test_mtime(storage, empty_file):
    storage.put(empty_file, "mtime.check")
    assert storage.mtime("mtime.check") == datetime.datetime.fromtimestamp(
        os.path.getmtime(os.path.join(storage.settings.dir, "mtime.check")))



# -*- coding: utf-8 -*-
"""
    :copyright: Copyright 2013-2014 by Łukasz Mierzwa
    :contact: l.mierzwa@gmail.com
"""


from __future__ import unicode_literals

import os
import logging

from upaas import commands


log = logging.getLogger(__name__)


def pack_tar(source, archive_path, timeout=None):
    """
    Pack files at given directory into tar archive.

    :param source: Directory which content should be packed.
    :param archive_path: Path at which tar archive file will be created.
    :param timeout: Timeout in seconds.
    """
    def _cleanup(archive_path):
        try:
            log.debug("Removing incomplete tar archive '%s' if "
                      "present" % archive_path)
            os.remove(archive_path)
        except OSError:
            pass

    cmd = "tar -czpf %s *" % archive_path

    # check if pigz is installed
    try:
        commands.execute("pigz -V", timeout=10)
    except commands.CommandError:
        pass
    else:
        cmd = "tar --use-compress-program=pigz -cpf %s *" % archive_path
        log.info("Using pigz for parallel compression")

    try:
        commands.execute(cmd, timeout=timeout, cwd=source)
    except commands.CommandTimeout:
        log.error("Tar command was taking too long and it was killed")
        _cleanup(archive_path)
        return False
    except commands.CommandFailed:
        log.error("Tar command failed")
        _cleanup(archive_path)
        return False
    else:
        return True


def unpack_tar(archive_path, destination, timeout=None):
    """
    Unpack tar archive in destination directory.

    :param archive_path: Path to tar file.
    :param destination: Destination directory in which we will unpack tar file.
    :param timeout: Timeout in seconds.
    """
    try:
        commands.execute("tar -xzpf %s" % archive_path, timeout=timeout,
                         cwd=destination)
    except commands.CommandTimeout:
        log.error("Tar command was taking too long and it was killed")
        return False
    except commands.CommandFailed:
        log.error("Tar command failed")
        return False
    else:
        return True



# -*- coding: utf-8 -*-
"""
    :copyright: Copyright 2013-2014 by Łukasz Mierzwa
    :contact: l.mierzwa@gmail.com
"""


from __future__ import unicode_literals

import platform


def distro_name():
    return platform.dist()[0]


def distro_version():
    return platform.dist()[1]


def distro_codename():
    return platform.dist()[2]


def distro_arch():
    return platform.architecture()[0]


def distro_image_filename():
    # FIXME tar.gz is hardcoded?
    path = "%s-%s-%s.tar.gz" % (distro_name(), distro_version(), distro_arch())
    return path.replace('/', '-')



# -*- coding: utf-8 -*-
"""
    :copyright: Copyright 2013-2014 by Łukasz Mierzwa
    :contact: l.mierzwa@gmail.com
"""


from __future__ import unicode_literals

import os
import shutil
import logging
import datetime

from upaas.config import base
from upaas.storage.base import BaseStorage
from upaas.storage.exceptions import StorageError, FileNotFound


log = logging.getLogger(__name__)


class LocalStorage(BaseStorage):

    configuration_schema = {
        "dir": base.FSPathEntry(required=True, must_exist=True),
    }

    def _join_paths(self, remote_path):
        return os.path.join(self.settings.dir, remote_path.lstrip('/'))

    def get(self, remote_path, local_path):
        if not self.exists(remote_path):
            log.error("[GET] File not found: %s" % remote_path)
            raise FileNotFound("%s not found" % remote_path)

        log.info("[GET] Copying %s to %s" % (self._join_paths(remote_path),
                                             local_path))
        try:
            shutil.copy(self._join_paths(remote_path), local_path)
        except Exception as e:
            raise StorageError(e)

    def put(self, local_path, remote_path):
        log.info("[PUT] Copying %s to %s" % (local_path,
                                             self._join_paths(remote_path)))
        try:
            shutil.copy(local_path, self._join_paths(remote_path))
        except Exception as e:
            raise StorageError(e)

    def delete(self, remote_path):
        if not self.exists(remote_path):
            log.error("[DELETE] File not found: %s" % remote_path)
            raise FileNotFound("%s not found" % remote_path)
        try:
            os.remove(self._join_paths(remote_path))
            log.info("[DELETE] File deleted: "
                     "%s" % self._join_paths(remote_path))
        except Exception as e:
            raise StorageError(e)

    def exists(self, remote_path):
        return os.path.exists(self._join_paths(remote_path))

    def size(self, remote_path):
        return os.path.getsize(self._join_paths(remote_path))

    def mtime(self, remote_path):
        return datetime.datetime.fromtimestamp(os.path.getmtime(
            self._join_paths(remote_path)))



# -*- coding: utf-8 -*-
"""
    :copyright: Copyright 2013-2014 by Łukasz Mierzwa
    :contact: l.mierzwa@gmail.com
"""


from __future__ import unicode_literals

import os
import time
import signal
import shutil
import logging

from upaas.commands import execute
from upaas.utils import umount_filesystems

log = logging.getLogger(__name__)


def directory_pids(directory):
    """
    List pid of all processes running inside given directory.
    List will contain integers and will use ascending sorting.

    :param directory: Directory to scan for running processes.
    :type directory: str

    :returns: list of int -- [134, 245, 673, 964]
    """
    log.debug("Scanning for processes running in %s" % directory)
    ret = set()
    if os.path.exists(directory):
        (rcode, output) = execute('lsof -t +d %s' % directory,
                                  valid_retcodes=[0, 1])
        if output:
            for line in output:
                try:
                    ret.add(int(line))
                except ValueError:
                    log.debug("Could not convert PID value to int: "
                              "'%s'" % line)
            return sorted(list(ret))
    else:
        log.debug("No such directory: %s" % directory)
    return []


def is_pid_running(pid):
    """
    Check if we have running process with given PID.

    :param pid: PID of process to check.
    :type pid: int
    :returns: bool -- True if given PID is running, False otherwise.
    """
    try:
        os.kill(pid, 0)
    except OSError:
        return False
    else:
        return True


def get_pid_command(pid):
    """
    Get command line of process with given PID. If command line cannot be
    found, then None is returned.

    :param pid: PID of process to lookup.
    :type pid: int

    :returns: str or None
    """
    try:
        with open("/proc/%d/cmdline" % pid, 'r') as proc:
            ret = proc.read()
        return ret or None
    except Exception as e:
        log.debug("Exception during /proc file parsing: %s" % e)


def wait_for_pid(pid, kill_after=600):
    """
    Wait for process to die, if it takes more than *kill_after* seconds than
    kill this process. Process is killed by sending SIGTERM and waiting 10
    seconds, if it's alive after that time SIGKILL is sent.

    :param pid: PID of process we wait to die.
    :type pid: int

    :param kill_after: Number of seconds to wait for process to die, after this
                       time it will be killed. Default is 600 seconds.
    :type kill_after: int
    """
    elapsed = 0
    while True:
        if is_pid_running(pid):
            log.debug("Waiting for pid %s to terminate (%s seconds "
                      "elapsed)" % (pid, elapsed))
            time.sleep(1)
            elapsed += 1
            if elapsed >= kill_after:
                log.info("%s seconds elapsed, killing process %s" % (elapsed,
                                                                     pid))
                os.kill(pid, signal.SIGKILL)
                elapsed = 0
                kill_after = 10
        else:
            break


def kill_pid(pid, timeout=60):
    """
    Kill running process by sending SIGTERM. If this process does not terminate
    after *timeout* seconds than it will be killed with SIGKILL signal.

    :param timeout: Number of seconds to wait before sending SIGKILL.
    :type timeout: int
    """
    if pid == os.getpid():
        log.debug("%d is my own PID, will not kill" % pid)
        return
    cmdline = get_pid_command(pid)
    log.info("Sending SIGTERM to %s [%s]" % (pid, cmdline or 'N/A'))
    try:
        os.kill(pid, signal.SIGTERM)
    except OSError:
        log.debug("PID %s already died" % pid)
    else:
        wait_for_pid(pid, kill_after=timeout)


def kill_and_remove_dir(directory):
    """
    Kill all processes running inside given directory and remove it.
    Used mostly for removing temporary directories.
    """
    found = True
    while found:
        for pid in directory_pids(directory):
            kill_pid(pid)
        else:
            found = False

    try:
        umount_filesystems(directory)
    except Exception as e:
        log.error("Error while unmounting filesystem inside "
                  "package: %s" % e)
    else:
        log.info("Removing directory: %s" % directory)
        shutil.rmtree(directory.encode('utf-8'))



# -*- coding: utf-8 -*-
"""
    :copyright: Copyright 2013-2014 by Łukasz Mierzwa
    :contact: l.mierzwa@gmail.com
"""


from __future__ import unicode_literals

import os
import shutil
import tempfile

import pytest
from _pytest.monkeypatch import monkeypatch


@pytest.fixture(scope="function")
def empty_dir(request):
    directory = tempfile.mkdtemp(prefix="upaas_testdir_")

    def cleanup():
        shutil.rmtree(directory)
    request.addfinalizer(cleanup)

    return directory


@pytest.fixture(scope="function")
def empty_file(request):
    path = tempfile.mkstemp(prefix="upaas_testfile_")[1]

    def cleanup():
        os.remove(path)
    request.addfinalizer(cleanup)

    return path


@pytest.fixture(scope="function")
def builder_config(request):

    directory = tempfile.mkdtemp(prefix="upaas_testdir_")

    class BuilderConfig:

        class paths:
            workdir = '/tmp'
            apps = '/tmp'
            vassals = '/tmp'

        class storage:
            handler = 'upaas.storage.local.LocalStorage'
            settings = {'dir': directory}

        class bootstrap:
            timelimit = 10
            maxage = 999
            env = {'LANG': 'C.UTF-8'}
            commands = ['/bin/touch bootstrapped.txt']
            packages = ['fakepackage']

        class commands:
            timelimit = 10

            class install:
                env = {'FAKEKEY': 'FAKEVALUE'}
                cmd = '/bin/echo'

            class uninstall:
                env = {'FAKEKEY': 'FAKEVALUE'}
                cmd = '/bin/echo'

        class apps:
            uid = 1000
            gid = 1000
            home = '/tmp'

        interpreters = {
            'ruby': {
                'any': {
                    'env': {},
                    'packages': ['fakerubypackage'],
                },
                '1.8.7': {
                    'packages': ['fakerubyversionpackage'],
                    'actions': {
                        'setup': {
                            'interpreter': ['/bin/echo interpreter action']
                        }
                    }
                }
            }
        }

    def cleanup():
        shutil.rmtree(directory)
    request.addfinalizer(cleanup)

    return BuilderConfig()


@pytest.fixture(scope="function")
def mock_chroot(request):
    class MockChroot(object):
        def __init__(self, *args, **kwargs):
            pass

        def __enter__(self):
            pass

        def __exit__(self, type, value, traceback):
            pass

    mpatch = monkeypatch()
    mpatch.setattr('upaas.chroot.Chroot', MockChroot)
    mpatch.setattr('upaas.builder.builder.Chroot', MockChroot)
    request.addfinalizer(mpatch.undo)


@pytest.fixture(scope="function")
def mock_build_commands(request):
    from upaas.commands import execute as real_execute

    def mock_execute(cmd, *args, **kwargs):
        executable = cmd.split(' ')[0]
        if executable in ['git', 'chown', 'gem', 'rake', 'bundle']:
            return 0, []
        else:
            return real_execute(cmd, *args, **kwargs)

    mpatch = monkeypatch()
    mpatch.setattr('upaas.commands.execute', mock_execute)
    request.addfinalizer(mpatch.undo)



# -*- coding: utf-8 -*-
"""
    :copyright: Copyright 2013-2014 by Łukasz Mierzwa
    :contact: l.mierzwa@gmail.com
"""


from __future__ import unicode_literals

from upaas import commands

import pytest


def test_failed_command():
    with pytest.raises(commands.CommandFailed):
        commands.execute("non existing command")


def test_cwd_and_output(empty_dir):
    _, output = commands.execute("pwd", cwd=empty_dir)
    assert output == [empty_dir + "\n"]


def test_timeout():
    with pytest.raises(commands.CommandTimeout):
        commands.execute("sleep 2", timeout=1)


def test_return_code():
    rcode, _ = commands.execute("exit 123", valid_retcodes=[123])
    assert rcode == 123


def test_env_and_output():
    _, output = commands.execute("echo $MYENV", env={"MYENV": "MYVALUE"})
    assert output == ["MYVALUE\n"]



# -*- coding: utf-8 -*-
"""
    :copyright: Copyright 2013-2014 by Łukasz Mierzwa
    :contact: l.mierzwa@gmail.com
"""


from __future__ import unicode_literals

import os

import pytest

from upaas.config import base
from upaas.compat import unicode


class BasicConfig(base.Config):
    schema = {
        "required_string": base.StringEntry(required=True),
        "folder1": {
            "subfolder1": {
                "required_int": base.IntegerEntry(required=True)
            },
            "optional_int": base.IntegerEntry(),
        },
        "chained": {
            "included": base.BooleanEntry()
        }
    }


class InvalidConfig(base.Config):
    schema = {
        "valid": base.IntegerEntry(),
        "invalid": int(),
    }


class ListConfig(base.Config):
    schema = {
        "mylist": base.ListEntry(value_type=int)
    }


class ListConfigNoType(base.Config):
    schema = {
        "mylist": base.ListEntry()
    }


class DictConfig(base.Config):
    schema = {
        "mydict": base.DictEntry(value_type=unicode)
    }


class DictConfigNoType(base.Config):
    schema = {
        "mydict": base.DictEntry()
    }


class BoolConfig(base.Config):
    schema = {
        "mybool_false": base.BooleanEntry(),
        "mybool_true": base.BooleanEntry(),
        "mybool_missing": base.BooleanEntry(),
    }


class IntConfig(base.Config):
    schema = {
        "int_with_min": base.IntegerEntry(min_value=10),
        "int_with_max": base.IntegerEntry(max_value=10),
        "int_with_min_max": base.IntegerEntry(min_value=10, max_value=20),
        "int_with_single_value": base.IntegerEntry(min_value=5, max_value=5),
    }


def test_empty():
    with pytest.raises(base.ConfigurationError):
        BasicConfig({})


def test_required_entries_from_dict():
    cfg = BasicConfig({
        "required_string": "abc",
        "folder1": {
            "subfolder1": {
                "required_int": 123
            }
        }
    })
    assert cfg.required_string == "abc"
    assert cfg.folder1.subfolder1.required_int == 123
    with pytest.raises(AttributeError):
        print((cfg.folder1.optional_int))


def test_required_entries_from_file():
    cfg = BasicConfig.from_file(os.path.join(os.path.dirname(__file__),
                                             "test_config.yml"))
    assert cfg.required_string == "abc"
    assert cfg.folder1.subfolder1.required_int == 123
    with pytest.raises(AttributeError):
        print((cfg.folder1.optional_int))
    assert cfg.chained.included is True


def test_loading_from_missing_file():
    with pytest.raises(base.ConfigurationError):
        BasicConfig.from_file("/there/is/no/such/file.yaml")


def test_loading_from_string():
    options = "folder1:\n  subfolder1: {required_int: 123}\n" \
              "required_string: abc\n"
    cfg = BasicConfig.from_string(options)
    assert cfg.dump_string() == options


def test_dump():
    options = {
        "chained": {
            "included": True
        },
        "required_string": "abc",
        "folder1": {
            "subfolder1": {
                "required_int": 123
            }
        }
    }
    cfg = BasicConfig(options)
    assert cfg.dump() == options


def test_dump_string():
    options = {
        "required_string": "abc",
        "folder1": {
            "subfolder1": {
                "required_int": 123
            }
        }
    }
    cfg = BasicConfig(options)
    assert cfg.dump_string() == "folder1:\n  subfolder1: " \
                                "{required_int: 123}\nrequired_string: abc\n"


def test_default_value():
    class LocalConfig(base.Config):
        schema = {"myitem": base.StringEntry(default="default value")}

    cfg = LocalConfig({})
    assert cfg.myitem == "default value"


def test_list_entry_valid():
    cfg = ListConfig({"mylist": [1, 2, 3, 4]})
    assert cfg.mylist == [1, 2, 3, 4]

    cfg = ListConfig({})
    assert cfg.mylist == []


def test_list_entry_invalid():
    with pytest.raises(base.ConfigurationError):
        ListConfig({"mylist": {"a": 1}})

    with pytest.raises(base.ConfigurationError):
        ListConfig({"mylist": ["not an int", 1]})


def test_list_entry_no_type():
    cfg = ListConfigNoType({"mylist": [1, 'a', '', 4]})
    assert cfg.mylist == [1, 'a', '', 4]

    cfg = ListConfigNoType({})
    assert cfg.mylist == []


def test_dict_entry_valid():
    cfg = DictConfig({"mydict": {"keya": "valuea", "keyb": "valueb"}})
    assert cfg.mydict == {"keya": "valuea", "keyb": "valueb"}

    cfg = DictConfig({})
    assert cfg.mydict == {}


def test_dict_entry_invalid():
    with pytest.raises(base.ConfigurationError):
        DictConfig({"mydict": [1]})

    with pytest.raises(base.ConfigurationError):
        DictConfig({"mydict": {"not an unicode": 1}})


def test_dict_entry_no_type():
    cfg = DictConfigNoType({"mydict": {"keya": "valuea", "keyb": 1}})
    assert cfg.mydict == {"keya": "valuea", "keyb": 1}

    cfg = DictConfigNoType({})
    assert cfg.mydict == {}


def test_init_config_invalid_content():
    with pytest.raises(base.ConfigurationError):
        BasicConfig(123)


def test_init_config_invalid_schema():
    cfg = InvalidConfig({"valid": 1, "invalid": 2})
    assert cfg.valid == 1
    with pytest.raises(AttributeError):
        cfg.invalid == 2


def test_load_config_using_path():
    cfg = base.load_config(BasicConfig, 'test_config.yml',
                           directories=[os.path.dirname(__file__)])
    assert cfg is not None
    assert cfg.required_string == "abc"
    assert cfg.folder1.subfolder1.required_int == 123


def test_load_config_using_env():
    os.environ['UPAAS_CONFIG_DIR'] = os.path.dirname(__file__)
    cfg = base.load_config(BasicConfig, 'test_config.yml')
    assert cfg is not None
    assert cfg.required_string == "abc"
    assert cfg.folder1.subfolder1.required_int == 123


def test_load_config_invalid():
    cfg = base.load_config(BasicConfig, os.path.basename(__file__),
                           directories=[os.path.dirname(__file__)])
    assert cfg is None


def test_load_config_missing():
    cfg = base.load_config(BasicConfig, 'non existing file',
                           directories=['/non/existing/dir'])
    assert cfg is None


def test_bool_entry():
    cfg = BoolConfig({"mybool_false": False, "mybool_true": True})
    assert cfg.mybool_false is False
    assert cfg.mybool_true is True

    with pytest.raises(AttributeError):
        _ = cfg.mybool_missing


def test_bool_entry_invalid():
    with pytest.raises(base.ConfigurationError):
        BoolConfig({"mybool_false": 1})
    with pytest.raises(base.ConfigurationError):
        BoolConfig({"mybool_false": 'a'})


def test_integer_invalid_schema():
    with pytest.raises(ValueError):
        class InvalidIntConfig(base.Config):
            schema = {
                "int": base.IntegerEntry(min_value=2, max_value=1),
            }


def test_integer_min_value_valid():
    cfg = IntConfig({"int_with_min": 10})
    assert cfg.int_with_min == 10

    cfg = IntConfig({"int_with_min": 20})
    assert cfg.int_with_min == 20


def test_integer_min_value_invalid():
    with pytest.raises(base.ConfigurationError):
        _ = IntConfig({"int_with_min": 5})


def test_integer_max_value_valid():
    cfg = IntConfig({"int_with_max": 1})
    assert cfg.int_with_max == 1

    cfg = IntConfig({"int_with_max": 10})
    assert cfg.int_with_max == 10


def test_integer_max_value_invalid():
    with pytest.raises(base.ConfigurationError):
        _ = IntConfig({"int_with_max": 15})


def test_integer_min_max_value_valid():
    cfg = IntConfig({"int_with_min_max": 10})
    assert cfg.int_with_min_max == 10

    cfg = IntConfig({"int_with_min_max": 20})
    assert cfg.int_with_min_max == 20


def test_integer_min_max_value_invalid():
    with pytest.raises(base.ConfigurationError):
        _ = IntConfig({"int_with_min_max": 9})
    with pytest.raises(base.ConfigurationError):
        _ = IntConfig({"int_with_min_max": 21})


def test_integer_single_value_valid():
    cfg = IntConfig({"int_with_single_value": 5})
    assert cfg.int_with_single_value == 5


def test_integer_single_value_invalid():
    with pytest.raises(base.ConfigurationError):
        _ = IntConfig({"int_with_single_value": 4})
    with pytest.raises(base.ConfigurationError):
        _ = IntConfig({"int_with_single_value": 6})



# -*- coding: utf-8 -*-
"""
    :copyright: Copyright 2013-2014 by Łukasz Mierzwa
    :contact: l.mierzwa@gmail.com
"""


class InvalidStorageConfiguration(Exception):
    pass


class StorageError(Exception):
    pass


class FileNotFound(StorageError):
    pass


class FileAlreadyExists(StorageError):
    pass



# -*- coding: utf-8 -*-
"""
    :copyright: Copyright 2013-2014 by Łukasz Mierzwa
    :contact: l.mierzwa@gmail.com
"""


from __future__ import unicode_literals

import os
import logging

import yaml
from yaml import Loader, SafeLoader, YAMLError

from upaas.compat import unicode


log = logging.getLogger(__name__)


# paths for searching config files
UPAAS_CONFIG_DIRS = ['.', '/etc/upaas']


def construct_yaml_str(self, node):
    """
    Override the default string handling function to always return unicode
    objects.

    Code from:
    http://stackoverflow.com/questions/2890146/how-to-force-pyyaml-to-load
    -strings-as-unicode-objects
    """
    return self.construct_scalar(node)
Loader.add_constructor('tag:yaml.org,2002:str', construct_yaml_str)
SafeLoader.add_constructor('tag:yaml.org,2002:str', construct_yaml_str)


def yaml_include(loader, node):
    """
    Adds !include statement for nesting yaml files.
    """
    path = node.value
    log.debug("Loading included configuration from %s" % path)
    if not os.path.exists(path):
        path = os.path.join(os.path.dirname(loader.stream.name), node.value)
    with open(path) as inputfile:
        return yaml.safe_load(inputfile)
yaml.SafeLoader.add_constructor("!include", yaml_include)


class ConfigurationError(Exception):
    """
    Raised if user provided configuration is invalid (missing or invalid
    required options).
    """
    pass


class ConfigEntry(object):

    default = None

    def __init__(self, required=False, default=None):
        self.required = required
        if default is not None:
            self.default = default

    @staticmethod
    def fail(msg):
        log.error(msg)
        raise ConfigurationError(msg)

    def validate(self, value):
        """
        Validate entry value, raise exception if invalid.
        """
        pass

    def clean(self, value):
        """
        Cleanup entry value if needed, returned value will be new entry value.
        """
        return value


class WildcardEntry(ConfigEntry):
    pass


class StringEntry(ConfigEntry):
    pass


class IntegerEntry(ConfigEntry):

    def __init__(self, min_value=None, max_value=None, *args, **kwargs):
        if min_value and max_value and min_value > max_value:
            msg = "Minimal value (%d) must be lower than maximum value " \
                  "(%d)!" % (min_value, max_value)
            log.error(msg)
            raise ValueError(msg)
        self.min_value = min_value
        self.max_value = max_value
        super(IntegerEntry, self).__init__(*args, **kwargs)

    def validate(self, value):
        if value is not None and not isinstance(value, int):
            self.fail("Value must be integer, %s "
                      "given" % value.__class__.__name__)
        if value is not None and self.min_value and value < self.min_value:
            self.fail("Value %d is too low, minimal value is %d" % (
                      value, self.min_value))
        if value is not None and self.max_value and value > self.max_value:
            self.fail("Value %d is too high, maximal value is %d" % (
                      value, self.max_value))


class BooleanEntry(ConfigEntry):

    def validate(self, value):
        if value is not None and not isinstance(value, bool):
            self.fail("Value must be a bool, %s "
                      "given" % value.__class__.__name__)


class ListEntry(ConfigEntry):

    default = []

    def __init__(self, value_type=None, **kwargs):
        self.value_type = value_type
        super(ListEntry, self).__init__(**kwargs)

    def validate(self, value):
        if value is None:
            return
        if not isinstance(value, list):
            self.fail("Value must be list, %s "
                      "given" % value.__class__.__name__)
        if not self.value_type:
            return
        for elem in value:
            if not isinstance(elem, self.value_type):
                self.fail("List element '%r' must be instance of '%s', '%s' "
                          "given" % (elem, self.value_type.__name__,
                                     elem.__class__.__name__))


class DictEntry(ConfigEntry):

    default = {}

    def __init__(self, value_type=None, **kwargs):
        self.value_type = value_type
        super(DictEntry, self).__init__(**kwargs)

    def validate(self, value):
        if value is None:
            return
        if not isinstance(value, dict):
            self.fail("Value must be dict, %s "
                      "given" % value.__class__.__name__)
        if not self.value_type:
            return
        for name, elem in list(value.items()):
            if not isinstance(elem, self.value_type):
                self.fail("Dict key '%r' value must be instance of '%s', '%s'"
                          " given" % (name, self.value_type.__name__,
                                      elem.__class__.__name__))


class FSPathEntry(ConfigEntry):
    def __init__(self, must_exist=False, **kwargs):
        self.must_exist = must_exist
        super(FSPathEntry, self).__init__(**kwargs)

    def validate(self, value):
        if self.must_exist and not os.path.exists(value):
            self.fail("Required path '%s' does not exits" % value)


class ScriptEntry(ConfigEntry):
    """
    Used for scripts that can be defined either as a list of strings or one big
    multi-line string. In case of string it will be converted into list.
    """

    def clean(self, value):
        if isinstance(value, unicode):
            return [value]
        return value


class Config(object):
    """
    Configuration object, it will parse all entries and check if they are valid
    using predefined schema.

    Schema is a dictionary where values are ConfigEntry instances describing
    possible options and its syntax:
    {
        "required_string": config.StringEntry(required=True),
        "folder1": {
            "subfolder1": {
                "required_int": config.IntegerEntry(required=True)
            },
            "optional_int": config.IntegerEntry(),
        }
    }
    """

    schema = {}

    @classmethod
    def from_file(cls, path):
        try:
            with open(path) as config_file:
                content = yaml.safe_load(config_file)
        except IOError as e:
            msg = "Can't open configuration file '%s': %s" % (path, e)
            log.error(msg)
            raise ConfigurationError(msg)
        except YAMLError as e:
            msg = "Can't parse yaml file '%s': %s" % (path, e)
            log.error(msg)
            raise ConfigurationError(msg)

        return cls(content)

    @classmethod
    def from_string(cls, string):
        return cls(yaml.safe_load(string))

    def __init__(self, content, _schema=None, name=''):
        if _schema:
            self.schema = _schema
        self.name = name

        log.debug("Parsing key '%s', settings %s, schema %s" % (self.name,
                                                                content,
                                                                self.schema))

        if not isinstance(content, (dict, type(None))):
            self.fail("Invalid configuration, expected dict but got "
                      "%s" % content.__class__.__name__)

        self.content = content
        self.entries = {}
        self.children = set()

        for key, value in list(self.schema.items()):
            if content is None and isinstance(value, ConfigEntry) \
                    and value.required:
                self.fail("Empty configuration")
            if isinstance(value, dict):
                setattr(self, key, Config(content.get(key, {}), _schema=value,
                                          name=self.child_name(key)))
                self.children.add(key)
            elif isinstance(value, WildcardEntry):
                self.entries[key] = (content or {}).get(key)
            elif isinstance(value, ConfigEntry):
                self.parse_entry(key, value, (content or {}).get(key))
            else:
                log.warning("Invalid configuration entry: "
                            "%s" % self.child_name(key))

    def __getattr__(self, item):
        if item in self.entries:
            return self.entries[item]
        raise AttributeError("%s not found" % item)

    def child_name(self, key):
        return '.'.join([_f for _f in [self.name, key] if _f])

    @staticmethod
    def fail(msg):
        log.error(msg)
        raise ConfigurationError(msg)

    def get(self, item, default=None):
        try:
            return self.__getattr__(item)
        except AttributeError:
            return default

    def dump(self):
        """
        Dump all entries as dictionary.
        """
        ret = {}
        ret.update(self.entries)
        for key in self.children:
            ret[key] = getattr(self, key).dump()
        return ret

    def dump_string(self):
        """
        Dump all entries as string.
        """
        return yaml.safe_dump(self.content)

    def parse_entry(self, name, entry_schema, value):
        """
        Parse and validate single configuration entry using schema.
        """
        log.debug("Parsing configuration entry '%s'" % self.child_name(name))
        if entry_schema.required and value is None and \
                entry_schema.default is None:
            self.fail("Missing required configuration entry: "
                      "%s" % self.child_name(name))
        log.debug("Cleaning configuration entry '%s'" % self.child_name(name))
        value = entry_schema.clean(value)
        log.debug("Validating configuration entry "
                  "'%s'" % self.child_name(name))
        try:
            entry_schema.validate(value)
        except ConfigurationError as e:
            self.fail("Configuration entry %s is invalid: %s" % (
                self.child_name(name), e))
        else:
            if value is not None:
                self.entries[name] = value
                log.debug("Configuration entry %s with value "
                          "'%s'" % (self.child_name(name), value))
            elif entry_schema.default is not None:
                log.debug("Configuration entry %s is missing, using default"
                          " value: %s" % (self.child_name(name),
                                          entry_schema.default))
                self.entries[name] = entry_schema.default


def load_config(cls, filename, directories=UPAAS_CONFIG_DIRS):
    """
    Try to load config, return None in case of errors.

    :param cls: Class to use for config validation and parsing.
    :filename: Filename to look for.
    """

    upaas_config = None

    paths = [os.path.join(p, filename) for p in directories]
    if os.environ.get('UPAAS_CONFIG_DIR'):
        log.info("Adding directory '%s' from UPAAS_CONFIG_DIR env "
                 "variable" % os.environ.get('UPAAS_CONFIG_DIR'))
        paths = [os.path.join(os.environ.get('UPAAS_CONFIG_DIR'),
                              filename)] + paths

    for path in paths:
        log.debug("Trying to load %s config file from %s" % (cls.__name__,
                                                             path))
        if os.path.isfile(path):
            log.info("Loading %s config file from %s" % (cls.__name__,
                                                         path))
            try:
                upaas_config = cls.from_file(path)
            except ConfigurationError:
                log.error("Invalid %s config file at %s" % (cls.__name__,
                                                            path))
                return None
            else:
                break

    if not upaas_config:
        log.error("No config file found for %s" % filename)

    return upaas_config



# -*- coding: utf-8 -*-
"""
    :copyright: Copyright 2013-2014 by Łukasz Mierzwa
    :contact: l.mierzwa@gmail.com
"""


from __future__ import unicode_literals

import os
import logging

from upaas.config import base
from upaas.compat import unicode


log = logging.getLogger(__name__)


class VCSLazyEntry(base.StringEntry):

    entry_name = None
    commands = {}

    def log_msg(self, msg):
        log.info("%s autodetect - %s" % (self.entry_name, msg))

    def detect(self):
        if os.path.isdir('.git'):
            self.log_msg("git detected, using '%s'" % self.commands['git'])
            return self.commands['git']
        elif os.path.isdir('.svn'):
            self.log_msg("svn detected using '%s'" % self.commands['svn'])
            return self.commands['svn']
        elif os.path.isdir('.bzr'):
            self.log_msg("bazaar detected using '%s'" % self.commands['bzr'])
            return self.commands['bzr']
        elif os.path.isdir('.hg'):
            self.log_msg("mercurial detected using '%s'" % self.commands['hg'])
            return self.commands['hg']
        else:
            self.log_msg("unknown repository type, using "
                         "'%s'" % self.commands['unknown'])
            return self.commands['unknown']

    default = detect


class VCSRevisionIDEntry(VCSLazyEntry):

    entry_name = 'repository.revision'

    commands = {
        'git': 'git log -1 --format=%H',
        'svn': "svn info --non-interactive --trust-server-cert "
               "| grep Revision | egrep -o '[0-9]+'",
        'bzr': 'bzr revno',
        'hg': 'hg id -i',
        'unknown': 'date +%s -r .',
    }


class VCSAuthorEntry(VCSLazyEntry):

    entry_name = 'repository.author'

    commands = {
        'git': "git log -1 --pretty='%aN <%aE>'",
        'svn': "svn info --non-interactive --trust-server-cert "
               "| grep 'Last Changed Author:' "
               "| sed s/'^Last Changed Author: '/''/",
        'bzr': "bzr log -l 1 | grep 'committer:' | sed s/'committer: '/''/",
        'hg': "hg log -r . --template='{author}'",
        'unknown': "echo 'no author information available'",
    }


class VCSDateEntry(VCSLazyEntry):

    entry_name = 'repository.date'

    commands = {
        'git': "git log -1 --pretty='%at'",
        'svn': "svn info --non-interactive --trust-server-cert "
               "| grep 'Last Changed Date:' "
               "| sed s/'^Last Changed Date: '/''/ "
               "| cut -d '(' -f 1",
        'bzr': "bzr version-info --custom --template '{date}'",
        'hg': "hg log -r . --template='{date}'",
        'unknown': 'date +%s -r .',
    }


class VCSDescriptionEntry(VCSLazyEntry):

    entry_name = 'repository.description'

    commands = {
        'git': "git log -1 --pretty='%B'",
        'svn': "svn log -l 1 --non-interactive --trust-server-cert "
               "| egrep -v '^\-+$' "
               "| tail -n +3",
        'bzr': "bzr log -l 1 | egrep -v '^-+$|^revno:|^committer:|"
               "^branch nick:|^timestamp:|^message:' | sed s/'^  '/''/",
        'hg': "hg log -r . --template='{desc}'",
        'unknown': "echo 'no description information available'",
    }


class VCSChangeLogEntry(VCSLazyEntry):

    entry_name = 'repository.description'

    commands = {
        'git': 'git log --no-merges --format=medium %old%..%new%',
        'svn': 'svn log -r%new%:%old% --non-interactive --trust-server-cert',
        'bzr': 'bzr log -r%old%..%new%',
        'hg': 'hg log -r %old%:%new%',
        'unknown': "echo 'no changelog information available'",
    }


class MetadataConfig(base.Config):

    schema = {
        "os": base.WildcardEntry(),
        "interpreter": {
            "type": base.StringEntry(required=True),
            "versions": base.ListEntry(unicode),
            "settings": base.DictEntry(value_type=unicode),
        },
        "repository": {
            "env": base.DictEntry(value_type=unicode),
            "clone": base.ScriptEntry(required=True),
            "update": base.ScriptEntry(required=True),
            "revision": {
                "id": VCSRevisionIDEntry(),
                "author": VCSAuthorEntry(),
                "date": VCSDateEntry(),
                "description": VCSDescriptionEntry(),
                "changelog": VCSChangeLogEntry(),
            }
        },
        "env": base.DictEntry(value_type=unicode),
        "actions": {
            "setup": {
                "before": base.ScriptEntry(),
                "main": base.ScriptEntry(),
                "after": base.ScriptEntry(),
            }
        },
        "files": base.DictEntry(value_type=unicode),
        "uwsgi": {
            "settings": base.ListEntry(value_type=unicode)
        },
        "cron": base.WildcardEntry(),
    }



# -*- coding: utf-8 -*-
"""
    :copyright: Copyright 2013-2014 by Łukasz Mierzwa
    :contact: l.mierzwa@gmail.com
"""


try:
    unicode = unicode
    basestring = basestring
except NameError:
    unicode = str
    basestring = (str, bytes)



# -*- coding: utf-8 -*-
"""
    :copyright: Copyright 2013-2014 by Łukasz Mierzwa
    :contact: l.mierzwa@gmail.com
"""


from __future__ import unicode_literals

import os
import shutil
import re
import logging

from upaas import commands


log = logging.getLogger(__name__)


def bytes_to_human(num):
    for x in ['bytes', 'KB', 'MB', 'GB', 'TB', 'PB']:
        if 1024.0 > num > -1024.0:
            return "%3.1f %s" % (num, x)
        num /= 1024.0
    return "%3.1f %s" % (num, 'EB')


def version_to_tuple(v):
    return tuple(map(int, (v.split("."))))


def version_tuple_to_string(version):
    return '.'.join(str(v) for v in version)


def version_fuzzy_compare(v1, v2):
    """
    Fuzzy version matching, checks if given 2 version strings are equal.
    It compares most major parts of each version.
    Example:
    5 == 5.4 => true
    5.3 == 5.4 => false
    5.1.4 == 5.1 => true
    """
    v1t = version_to_tuple(v1)
    v2t = version_to_tuple(v2)
    match_items = min(len(v1t), len(v2t))
    return v1t[0:match_items] == v2t[0:match_items]


def supported_versions(config, metadata):
    """
    Return list of versions from metadata that are supported localy.
    """
    valid_versions = {}
    for version in metadata.interpreter.versions:
        for iversion in config.interpreters[metadata.interpreter.type]:
            if iversion.lower() != 'any':
                if version_fuzzy_compare(version, iversion):
                    valid_versions[iversion] = version_to_tuple(iversion)
    return valid_versions


def select_best_version(config, metadata):
    """
    Return highest supported version or None.
    """
    valid_versions = supported_versions(config, metadata)
    if valid_versions:
        return version_tuple_to_string(
            sorted(list(valid_versions.values()), reverse=True)[0])


def rmdirs(*args):
    for directory in args:
        if os.path.isdir(directory):
            log.info("Removing directory '%s'" % directory)
            shutil.rmtree(directory)


def umount_filesystems(workdir, timeout=60):
    mounts = []
    if os.path.isfile("/proc/mounts"):
        with open("/proc/mounts") as mtab:
            for line in mtab:
                try:
                    mount = line.split()[1]
                except IndexError:
                    pass
                else:
                    if mount.startswith(workdir.rstrip('/') + '/'):
                        mounts.append(mount)
    for mount in mounts:
        log.info("Found mounted filesystem at '%s', unmounting" % mount)
        commands.execute("umount %s" % mount, timeout=timeout)


def backend_total_memory():
    """
    Local backend physical memory size in bytes
    """
    with open('/proc/meminfo') as meminfo:
        matched = re.search(r'^MemTotal:\s+(\d+)', meminfo.read())
        if matched:
            try:
                return int(matched.groups()[0]) * 1024
            except:
                return 0
    return 0



# -*- coding: utf-8 -*-
"""
    :copyright: Copyright 2013-2014 by Łukasz Mierzwa
    :contact: l.mierzwa@gmail.com
"""


from __future__ import unicode_literals

import os
import tempfile
import datetime
import logging

from timestring import Date

from upaas import distro

from upaas import commands
from upaas import tar
from upaas import utils
from upaas.checksum import calculate_file_sha256
from upaas.builder import exceptions
from upaas.chroot import Chroot
from upaas.storage.utils import find_storage_handler
from upaas.storage.exceptions import StorageError
from upaas.processes import kill_and_remove_dir


log = logging.getLogger(__name__)


class BuildResult:

    def __init__(self):
        # % progress
        self.progress = 0

        # filename of the package used to build this package
        self.parent = None

        # selected interpreter version
        self.interpreter_version = None

        # package filename
        self.filename = None
        # package checksum
        self.checksum = None
        # package size
        self.bytes = 0

        self.distro_name = distro.distro_name()
        self.distro_version = distro.distro_version()
        self.distro_arch = distro.distro_arch()

        # information about last commit in VCS (if available)
        self.vcs_revision = {}


class Builder(object):

    # TODO stage decorators
    # @stage('action_system', 95%)
    # time()

    builder_action_names = ["system"]
    interpreter_action_names = ["interpreter"]
    app_action_names = ["before", "main", "after"]
    finalize_action_names = ["finalize"]

    def __init__(self, builder_config, metadata):
        """
        :param builder_config: Builder configuration.
        :param metadata: Application metadata.
        """
        self.metadata = metadata
        self.config = builder_config
        self.envs = {}
        self.actions = {}
        self.os_packages = []

        self.interpreter_version = utils.select_best_version(self.config,
                                                             metadata)

        self.current_revision = None

    def user_error(self, msg):
        log.error(msg)
        raise exceptions.PackageUserError(msg)

    def system_error(self, msg):
        log.error(msg)
        raise exceptions.PackageSystemError(msg)

    def parse_actions(self, meta):
        """
        Parse and merge all config files (builder and app meta), then return
        final list of all actions to perform.
        """
        def _run_action(entry, name, ret):
            try:
                value = entry["actions"]["setup"][name]
                if value:
                    ret[name] = value
            except KeyError:
                pass
            else:
                log.debug("Got '%s' action" % name)

        ret = {}

        # builder and finalize actions are special, they cannot be declared
        # by app, only by builder config
        for name in self.builder_action_names + self.finalize_action_names:
            ret[name] = []
            _run_action(self.config.interpreters, name, ret)

        # interpreter action can also be declared only by builder config,
        # but it is taken from interpreter settings key
        for name in self.interpreter_action_names:
            ret[name] = []
            try:
                ret[name] = self.config.interpreters[meta.interpreter.type][
                    "any"]["actions"]["setup"][name]
            except KeyError:
                pass
            try:
                ret[name] = self.config.interpreters[meta.interpreter.type][
                    self.interpreter_version]["actions"]["setup"][name]
            except KeyError:
                pass

        for name in self.app_action_names:
            ret[name] = []

            # global actions
            _run_action(self.config.interpreters, name, ret)

            # interpreter actions
            for version in ["any"] + [self.interpreter_version]:
                try:
                    cfg = self.config.interpreters[meta.interpreter.type][
                        version]
                except KeyError:
                    pass
                else:
                    _run_action(cfg, name, ret)

            # app metadata actions
            try:
                value = meta.actions.setup.get(name)
                if value:
                    ret[name] = value
            except AttributeError:
                pass
            else:
                log.debug("Got '%s' action from app meta" % name)

        for name in self.builder_action_names + \
                self.interpreter_action_names + self.app_action_names + \
                self.finalize_action_names:
            actions = ret.get(name, [])
            log.info("Commands for '%s' action:" % name)
            for action in actions:
                for line in action.splitlines():
                    log.info("- %s" % line)

        return ret

    def parse_envs(self, meta):
        """
        Parse and merge all config files (builder and app meta), then return
        final all env variables used in actions.
        """
        ret = {}
        try:
            value = self.config.interpreters["env"]
            if value:
                ret.update(value)
                log.info("Got env variables for all interpreters: %s" % (
                    ", ".join(["%s=%s" % (k, v) for k, v in list(
                        value.items())])))
        except KeyError:
            pass
        for version in ["any"] + [self.interpreter_version]:
            try:
                value = self.config.interpreters[meta.interpreter.type][
                    version]["env"]
                if value:
                    ret.update(value)
            except KeyError:
                pass
            else:
                log.info("Got env variables from %s/%s: %s" % (
                    meta.interpreter.type, version,
                    ", ".join(["%s=%s" % (k, v) for k, v in list(
                        value.items())])))
                if meta.get("env"):
                    ret.update(meta.env)
                    log.info("Got env variables from app meta: " +
                             ", ".join(["%s=%s" % (k, v)
                                       for k, v in list(value.items())]))

        if ret:
            log.info("Final env variables:")
            for key, value in list(ret.items()):
                log.info("%s = %s" % (key, value))
        else:
            log.info("No env variables set")

        return ret

    def parse_packages(self, meta):
        """
        Parse and merge all config files (builder and app meta), then return
        final list of all packages to install.
        """
        ret = set()
        try:
            for pkg in self.config.interpreters["packages"]:
                ret.add(pkg)
                log.debug("Will install package '%s' from builder config "
                          "for all interpreters" % pkg)
        except KeyError:
            pass
        for version in ["any"] + [self.interpreter_version]:
            try:
                for pkg in self.config.interpreters[meta.interpreter.type][
                        version]["packages"]:
                    ret.add(pkg)
                    log.debug("Will install package '%s' from builder config "
                              "for interpreter version '%s'" % (pkg, version))
            except KeyError:
                pass
            try:
                for pkg in meta.os[distro.distro_name()]["packages"]:
                    ret.add(pkg)
                    log.debug("Will install package '%s' from metadata for "
                              "distribution %s" % (pkg, distro.distro_name()))
            except (KeyError, TypeError):
                pass
            try:
                for pkg in meta.os[distro.distro_name()][
                        distro.distro_version()]["packages"]:
                    ret.add(pkg)
                    log.debug("Will install package '%s' from metadata for "
                              "distribution %s, version "
                              "%s" % (pkg, distro.distro_name(),
                                      distro.distro_version()))
            except (KeyError, TypeError):
                pass
        return ret

    def build_package(self, system_filename=None, interpreter_version=None,
                      current_revision=None):
        """
        Build a package

        :param system_filename: Use given file as base system, if None empty
                                system image will be used.
        :param interpreter_version: Use specific interpreter version, only used
                                    for fresh packages.
        :param current_revision: VCS revision id from current package.
        """
        if interpreter_version:
            self.interpreter_version = interpreter_version
            log.info("Using forced interpreter version: "
                     "%s" % interpreter_version)

        self.storage = find_storage_handler(self.config.storage.handler,
                                            self.config.storage.settings)
        if system_filename and self.storage.exists(system_filename):
            log.info("Starting package build using package "
                     "%s" % system_filename)
            if current_revision:
                log.info("VCS revision from current package: "
                         "%s" % current_revision)
                self.current_revision = current_revision
        else:
            if system_filename:
                log.warning("Requested base package file not found, using "
                            "empty system image")
            self.envs['UPAAS_FRESH_PACKAGE'] = 'true'
            system_filename = None
            log.info("Starting package build using empty system image")
            if not self.has_valid_os_image():
                try:
                    self.bootstrap_os()
                except exceptions.OSBootstrapError as e:
                    self.system_error("Error during os bootstrap: %s" % e)
                except StorageError as e:
                    self.system_error("Error during uploading OS image: "
                                      "%s" % e)

        if not self.interpreter_version:
            self.user_error("Unsupported interpreter version")

        self.actions.update(self.parse_actions(self.metadata))
        self.envs.update(self.parse_envs(self.metadata))
        self.os_packages += self.parse_packages(self.metadata)

        result = BuildResult()
        result.parent = system_filename
        result.interpreter_version = self.interpreter_version

        # directory is encoded into string to prevent unicode errors
        directory = tempfile.mkdtemp(dir=self.config.paths.workdir,
                                     prefix="upaas_package_")
        workdir = os.path.join(directory, "workdir")
        chroot_homedir = self.config.apps.home
        os.mkdir(workdir, 0o755)
        log.info("Working directory created at '%s'" % workdir)
        self.envs['HOME'] = chroot_homedir

        if not self.unpack_os(directory, workdir,
                              system_filename=system_filename):
            kill_and_remove_dir(directory)
            self.system_error("Unpacking OS image failed")
        log.info("OS image unpacked")
        result.progress = 10
        yield result

        log.info("Using interpreter %s, version %s" % (
            self.metadata.interpreter.type, self.interpreter_version))

        if not self.run_actions(self.builder_action_names, workdir):
            kill_and_remove_dir(directory)
            self.system_error("System actions failed")
        log.info("All builder actions executed")
        result.progress = 20
        yield result

        if not self.install_packages(workdir, self.os_packages):
            kill_and_remove_dir(directory)
            self.user_error("Failed to install OS packages")
        log.info("All packages installed")
        result.progress = 35
        yield result

        if not self.run_actions(self.interpreter_action_names, workdir, '/'):
            kill_and_remove_dir(directory)
            self.system_error("Interpreter actions failed")
        log.info("All interpreter actions executed")
        result.progress = 40
        yield result

        # TODO if building fails up to this point, then we can try retry it
        # on another builder (for a limited number of times)

        if system_filename:
            if not self.update(workdir, chroot_homedir):
                kill_and_remove_dir(directory)
                self.user_error("Updating repository failed")
        else:
            if not self.clone(workdir, chroot_homedir):
                kill_and_remove_dir(directory)
                self.user_error("Cloning repository failed")
        log.info("Application repository ready")
        result.progress = 45
        yield result

        result.vcs_revision = self.vcs_info(workdir, chroot_homedir)
        result.progress = 46
        yield result

        if not self.write_files(workdir, chroot_homedir):
            kill_and_remove_dir(directory)
            self.user_error("Creating files from metadata failed")
        log.info("Created all files from metadata")
        result.progress = 49
        yield result

        if not self.run_actions(self.app_action_names, workdir,
                                chroot_homedir):
            self.user_error("Application actions failed")
            kill_and_remove_dir(directory)
        log.info("All application actions executed")
        result.progress = 85
        yield result

        if not self.run_actions(self.finalize_action_names, workdir, '/'):
            kill_and_remove_dir(directory)
            self.system_error("Finalize actions failed")
        log.info("All final actions executed")
        result.progress = 88
        yield result

        if not self.chown_app_dir(workdir, chroot_homedir):
            kill_and_remove_dir(directory)
            self.system_error("Setting file ownership failed")
        log.info("Owner of application directory updated")
        result.progress = 89
        yield result

        if not self.umount_filesystems(workdir):
            kill_and_remove_dir(directory)
            self.system_error("Failed to unmount filesystems")
        result.progress = 90
        yield result

        package_path = os.path.join(directory, "package")
        if not tar.pack_tar(workdir, package_path):
            kill_and_remove_dir(directory)
            self.system_error("Creating package file failed")
        result.bytes = os.path.getsize(package_path)
        log.info("Application package created, "
                 "%s" % utils.bytes_to_human(result.bytes))
        result.progress = 93
        yield result

        checksum = calculate_file_sha256(package_path)
        log.info("Package checksum: %s" % checksum)
        result.progress = 96
        yield result

        try:
            self.storage.put(package_path, checksum)
        except StorageError as e:
            kill_and_remove_dir(directory)
            self.system_error("Package upload failed: %s" % e)

        kill_and_remove_dir(directory)

        result.progress = 100
        result.filename = checksum
        result.checksum = checksum
        yield result

    def unpack_os(self, directory, workdir, system_filename=None):
        empty_os_image = False
        if not system_filename:
            system_filename = distro.distro_image_filename()
            empty_os_image = True

        os_image_path = os.path.join(directory, "os.image")
        log.info("Fetching OS image '%s'" % system_filename)
        try:
            self.storage.get(system_filename, os_image_path)
        except StorageError:
            log.error("Storage error while fetching OS image")
            return False
        else:
            log.info("Unpacking OS image")
            if not tar.unpack_tar(os_image_path, workdir):
                log.error("Error while unpacking OS image to '%s'" % workdir)
                return False
        # verify if os is working
        log.info("Checking if OS image is working (will execute /bin/true)")
        try:
            with Chroot(workdir):
                commands.execute('/bin/true',
                                 timeout=self.config.commands.timelimit,
                                 output_loglevel=logging.INFO)
        except Exception as e:
            log.error("Broken OS image! /bin/true failed: %s" % e)
            if empty_os_image:
                try:
                    self.storage.delete(system_filename)
                except StorageError as e:
                    log.error("Storage error while deleting OS image: %s" % e)
                else:
                    log.info("Deleted broken OS image from storage")
                    self.bootstrap_os()
                    return self.unpack_os(directory, workdir,
                                          system_filename=system_filename)
            return False
        else:
            return True

    def install_packages(self, workdir, packages):
        with Chroot(workdir):
            for name in packages:
                cmd = self.config.commands.install.cmd.replace("%package%",
                                                               name)
                try:
                    commands.execute(cmd,
                                     timeout=self.config.commands.timelimit,
                                     env=self.config.commands.install.env,
                                     output_loglevel=logging.INFO,
                                     strip_envs=True)
                except commands.CommandTimeout:
                    log.error("Installing package '%s' is taking to long, "
                              "aborting" % name)
                    return False
                except commands.CommandFailed:
                    log.error("Installing package '%s' failed" % name)
                    return False
        return True

    def clone(self, workdir, homedir):
        log.info("Updating repository to '%s'" % homedir)
        with Chroot(workdir):
            for cmd in self.metadata.repository.clone:
                cmd = cmd.replace("%destination%", homedir)
                try:
                    commands.execute(cmd,
                                     timeout=self.config.commands.timelimit,
                                     env=self.metadata.repository.env,
                                     output_loglevel=logging.INFO,
                                     strip_envs=True)
                except commands.CommandTimeout:
                    log.error("Command is taking too long, aborting")
                    return False
                except commands.CommandFailed:
                    log.error("Command failed")
                    return False
        return True

    def vcs_info(self, workdir, homedir):
        ret = {}
        log.info("Extracting information about last commit")
        with Chroot(workdir, workdir=homedir):
            def vcs_cmd(name, cmd, replace=None):
                name = 'repository.revision.%s' % name
                if hasattr(cmd, '__call__'):
                    cmd = cmd()
                if replace:
                    for (rold, rnew) in replace:
                        cmd = cmd.replace(rold, rnew)
                env = self.metadata.repository.env.copy()
                env['LANG'] = 'C.UTF-8'
                env['LC_ALL'] = 'C.UTF-8'
                try:
                    _, output = commands.execute(
                        cmd, timeout=self.config.commands.timelimit, env=env,
                        output_loglevel=logging.INFO, strip_envs=True)
                except commands.CommandTimeout:
                    log.error("%s command is taking too long, aborting" % name)
                except commands.CommandFailed:
                    log.error("%s command failed" % name)
                else:
                    return ''.encode('utf-8').join(output).rstrip(
                        '\n'.encode('utf-8'))

            ret = {
                'id': vcs_cmd('id', self.metadata.repository.revision.id),
                'author': vcs_cmd('author',
                                  self.metadata.repository.revision.author),
                'date': vcs_cmd('date',
                                self.metadata.repository.revision.date),
                'description': vcs_cmd(
                    'description',
                    self.metadata.repository.revision.description),
            }

            if self.current_revision and self.current_revision != ret['id']:
                ret['changelog'] = vcs_cmd(
                    'changelog',
                    self.metadata.repository.revision.changelog,
                    replace=[
                        ('%old%', self.current_revision.rstrip('\n')),
                        ('%new%', ret['id'].rstrip('\n')),
                    ])

        if 'date' in ret:
            try:
                ret['date'] = Date(ret['date']).date
            except ValueError:
                log.warning("Can't convert '%s' to date" % ret['date'])
                del ret['date']

        return ret

    def update(self, workdir, homedir):
        log.info("Updating repository in '%s'" % homedir)
        with Chroot(workdir, workdir=homedir):
            for cmd in self.metadata.repository.update:
                cmd = cmd.replace("%destination%", homedir)
                try:
                    commands.execute(cmd,
                                     timeout=self.config.commands.timelimit,
                                     env=self.metadata.repository.env,
                                     output_loglevel=logging.INFO,
                                     strip_envs=True)
                except commands.CommandTimeout:
                    log.error("Command is taking too long, aborting")
                    return False
                except commands.CommandFailed:
                    log.error("Command failed")
                    return False
        return True

    def run_actions(self, actions, workdir, homedir='/'):
        for name in actions:
            log.info("Executing '%s' setup actions" % name)
            for cmd in self.actions[name]:
                with Chroot(workdir, workdir=homedir):
                    try:
                        commands.execute(
                            cmd, timeout=self.config.commands.timelimit,
                            env=self.envs, output_loglevel=logging.INFO,
                            strip_envs=True)
                    except commands.CommandTimeout:
                        log.error("Command is taking too long to execute, "
                                  "aborting")
                        return False
                    except commands.CommandFailed as e:
                        log.error("Execution failed: %s" % e)
                        return False
        return True

    def chown_app_dir(self, workdir, homedir):
        cmd = "chown -R %s:%s %s" % (self.config.apps.uid,
                                     self.config.apps.gid, homedir)
        with Chroot(workdir):
            try:
                commands.execute(cmd, timeout=self.config.commands.timelimit,
                                 output_loglevel=logging.INFO, strip_envs=True)
            except commands.CommandTimeout:
                log.error("chown is taking too long to execute, aborting")
                return False
            except commands.CommandFailed:
                log.error("chown failed")
                return False
        return True

    def has_valid_os_image(self):
        """
        Check if OS image exists and is fresh enough.
        """
        if not self.storage.exists(distro.distro_image_filename()):
            return False

        os_mtime = self.storage.mtime(distro.distro_image_filename())
        delta = datetime.datetime.now() - os_mtime
        if delta > datetime.timedelta(days=self.config.bootstrap.maxage):
            log.info("OS image is too old (%d days)" % delta.days)
            self.storage.delete(distro.distro_image_filename())
            return False

        return True

    def bootstrap_os(self):
        """
        Bootstrap base OS image.
        """
        log.info("Bootstrapping new OS image")

        # directory is encoded into string to prevent unicode errors
        directory = tempfile.mkdtemp(dir=self.config.paths.workdir,
                                     prefix="upaas_bootstrap_")
        log.debug("Created temporary directory for bootstrap at "
                  "'%s'" % directory)

        for cmd in self.config.bootstrap.commands:
            cmd = cmd.replace("%workdir%", directory)
            try:
                commands.execute(cmd, timeout=self.config.bootstrap.timelimit,
                                 cwd=directory, env=self.config.bootstrap.env,
                                 strip_envs=True)
            except commands.CommandTimeout as e:
                log.error("Bootstrap was taking too long and it was killed")
                kill_and_remove_dir(directory)
                raise exceptions.OSBootstrapError(e)
            except commands.CommandFailed as e:
                log.error("Bootstrap command failed")
                kill_and_remove_dir(directory)
                raise exceptions.OSBootstrapError(e)
        log.info("All commands completed, installing packages")

        self.install_packages(directory, self.config.bootstrap.packages)
        log.info("Bootstrap done, packing image")

        archive_path = os.path.join(directory, "image.tar.gz")
        if not tar.pack_tar(directory, archive_path,
                            timeout=self.config.bootstrap.timelimit):
            kill_and_remove_dir(directory)
            raise exceptions.OSBootstrapError("Tar error")
        else:
            log.info("Image packed, uploading")

        try:
            self.storage.put(archive_path, distro.distro_image_filename())
        except Exception as e:
            log.error("Upload failed: %s" % e)
            raise

        log.info("Image uploaded")
        kill_and_remove_dir(directory)
        log.info("All done")

    def write_files(self, workdir, chroot_homedir):
        """
        Create all files specified in metadata 'files' section.
        """
        with Chroot(workdir, workdir=chroot_homedir):
            for path, content in list(self.metadata.files.items()):
                basedir = os.path.dirname(path)
                if basedir and not os.path.exists(basedir):
                    try:
                        os.makedirs(basedir)
                    except Exception as e:
                        log.error("Can't create base directory (%s): %s" % (
                            basedir, e))
                        return False
                log.info("Creating metadata file: %s" % path)
                try:
                    with open(path, 'w') as out:
                        out.write(content)
                except Exception as e:
                    log.error("Can't write to '%s': %s" % (path, e))
                    return False
        return True

    def umount_filesystems(self, workdir):
        try:
            utils.umount_filesystems(workdir,
                                     timeout=self.config.bootstrap.timelimit)
        except commands.CommandTimeout as e:
            log.error("Can't umount filesystem, timeout reached")
            return False
        except commands.CommandFailed as e:
            log.error("Failed to umount filesystem: %s" % e)
            return False
        return True


class OSBuilder(Builder):

    def __init__(self, builder_config):
        """
        :param builder_config: Builder configuration.
        """
        self.config = builder_config
        self.storage = find_storage_handler(self.config.storage.handler,
                                            self.config.storage.settings)



# -*- coding: utf-8 -*-
"""
    :copyright: Copyright 2013-2014 by Łukasz Mierzwa
    :contact: l.mierzwa@gmail.com
"""


from __future__ import unicode_literals

import os
import subprocess
import signal
import logging


log = logging.getLogger(__name__)


SAFE_ENVS = [
    'HOME',
    'LOGNAME',
    'USER',
    'TERM',
    'PATH',
    'LANG',
    'LC_ALL',
    'TZ',
    'SHELL',
    'PWD',
]


class CommandError(Exception):
    """
    Generic command execution error.
    """
    pass


class CommandTimeout(CommandError):
    """
    Command was executing longer than allowed.
    """
    pass


class CommandFailed(CommandError):
    """
    Command exited with return code indicating error.
    """
    pass


def execute(cmd, timeout=None, cwd=None, output_loglevel=logging.DEBUG, env={},
            valid_retcodes=[0], strip_envs=False):
    """
    Execute given command in shell.

    :param timeout: Maximum time (in seconds) command can take to execute, if
                    it takes longer it will be killed. No timeout if None.
    :param cwd: If provided chdir() to this path before executing.
    :param output_loglevel: Logging level at which command output will be
                            logged.
    :param env: Dictionary with environment variables for this command.
    :param valid_retcodes: List of return codes that can be returned by this
                           command. Other return codes will be interpreted as
                           error and exception will be raised.
    :param strip_envs: If True all unsafe env variables will be removed
                       before executing command.
    :returns: tuple -- (return code, output as list of strings)
    """
    def _alarm_handler(signum, frame):
        raise CommandTimeout("Command timeout reached")

    def _cleanup(workdir, original_env):
        if workdir:
            log.debug("Switching back to workdir at '%s'" % workdir)
            os.chdir(workdir)
        for key, value in list(original_env.items()):
            if value:
                log.debug("Restoring original ENV variable %s=%s" % (key,
                                                                     value))
                os.environ[key] = value
            else:
                log.debug("Deleting extra ENV variable %s" % key)
                try:
                    del os.environ[key]
                except KeyError:
                    pass

    log.info("Executing command: %s" % cmd, extra={"force_flush": True})

    wd = None
    if cwd:
        wd = os.getcwd()
        log.info("Changing working directory to '%s'" % cwd)
        os.chdir(cwd)

    original_env = {}
    if strip_envs:
        for ename, evalue in list(os.environ.items()):
            if ename not in (list(env.keys()) + SAFE_ENVS):
                log.debug("Removing unsafe ENV variable %s=%s" % (ename,
                                                                  evalue))
                original_env[ename] = evalue
                del os.environ[ename]

    for ename, evalue in list(env.items()):
        log.debug("Setting ENV variable %s=%s" % (ename, evalue))
        orgval = os.environ.get(ename)
        log.debug("Saving original ENV value %s=%s" % (ename, orgval))
        original_env[ename] = orgval
        os.environ[ename] = evalue

    if timeout:
        signal.signal(signal.SIGALRM, _alarm_handler)
        signal.alarm(timeout)
        log.debug("Timeout for command is %d seconds" % timeout)

    output = []
    log.debug("Running ...")
    p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT,
                         shell=True)
    try:
        while True:
            retcode = p.poll()
            line = p.stdout.readline().decode('utf-8')
            if line:
                output.append(line)
                log.log(output_loglevel, line.rstrip(os.linesep))
            if retcode is not None:
                break
    except CommandTimeout:
        os.kill(p.pid, signal.SIGKILL)
        _cleanup(wd, original_env)
        raise CommandTimeout("Command timeout reached")
    except KeyboardInterrupt as e:
        if timeout:
            signal.alarm(0)
        os.kill(p.pid, signal.SIGKILL)
        _cleanup(wd, original_env)
        raise CommandFailed(e)
    except Exception:
        signal.alarm(0)
        raise

    if timeout:
        signal.alarm(0)

    _cleanup(wd, original_env)

    if retcode not in valid_retcodes:
        msg = "Command failed with status %d" % retcode
        log.error(msg)
        raise CommandFailed(msg)

    return retcode, output



# -*- coding: utf-8 -*-
"""
    :copyright: Copyright 2013-2014 by Łukasz Mierzwa
    :contact: l.mierzwa@gmail.com
"""


from __future__ import unicode_literals

import logging

from upaas.config.base import ConfigurationError

log = logging.getLogger(__name__)


def import_handler(module_name, cls_name):
    """
    Import class from module by name.
    """
    mod = __import__(module_name)
    components = module_name.split('.')
    for comp in components[1:]:
        mod = getattr(mod, comp)
    return getattr(mod, cls_name)


def find_storage_handler(name, settings=None):
    """
    Will try to find storage handler class user has set in configuration,
    create instance of it and return that instance.
    """
    storage_module = ".".join(name.split(".")[0:len(name.split(".")) - 1])
    storage_class = name.split(".")[len(name.split(".")) - 1]
    log.debug("Trying to import '%s' from '%s'" % (storage_class,
                                                   storage_module))
    try:
        storage_handler = import_handler(storage_module, storage_class)
    except ImportError:
        msg = "Storage handler '%s' could not be loaded" % name
        log.error(msg)
        raise ConfigurationError(msg)
    else:
        log.debug("Loaded storage handler '%s', initializing with settings: "
                  "%s" % (name, (settings or {}).keys()))
        try:
            return storage_handler(settings)
        except ConfigurationError:
            msg = "Storage handler failed to initialize with given " \
                  "configuration"
            log.error(msg)
            raise ConfigurationError(msg)



# -*- coding: utf-8 -*-
"""
    :copyright: Copyright 2013-2014 by Łukasz Mierzwa
    :contact: l.mierzwa@gmail.com
"""


class InvalidConfiguration(Exception):
    """
    Raised if upaas-builder configuration is invalid.
    """
    pass


class BuildError(Exception):
    """
    General class for catching any build error.
    """
    pass


class OSBootstrapError(BuildError):
    """
    Raised in case of errors during os image bootstraping.
    """
    pass


class PackageSystemError(BuildError):
    """
    Raised in case of system errors during package build. This does not cover
    errors caused by package configuration or any commands executed by package
    itself, only errors independent from package (os bootstrap error for
    example)
    """
    pass


class PackageUserError(BuildError):
    """
    Raised when executing package specific actions.
    """
    pass



# -*- coding: utf-8 -*-
"""
    :copyright: Copyright 2013-2014 by Łukasz Mierzwa
    :contact: l.mierzwa@gmail.com
"""


from hashlib import sha256


def calculate_file_sha256(path):
    hasher = sha256()
    with open(path, "rb") as sfile:
        while True:
            data = sfile.read(4096)
            if not data:
                break
            hasher.update(data)
    return hasher.hexdigest()


def calculate_string_sha256(content):
    hasher = sha256()
    hasher.update(content)
    return hasher.hexdigest()



# -*- coding: utf-8 -*-
"""
    :copyright: Copyright 2013-2014 by Łukasz Mierzwa
    :contact: l.mierzwa@gmail.com
"""


from __future__ import unicode_literals

import logging

from pymongo import MongoClient

from gridfs import GridFS, NoFile

from upaas.config import base
from upaas.storage.base import BaseStorage
from upaas.storage.exceptions import StorageError, FileNotFound,\
    FileAlreadyExists


log = logging.getLogger(__name__)


class MongoDBStorage(BaseStorage):

    configuration_schema = {
        "host": base.StringEntry(default="localhost"),
        "port": base.IntegerEntry(default=27017),
        "database": base.StringEntry(default="upaas-storage"),
        "username": base.StringEntry(),
        "password": base.StringEntry(),
    }

    def connect(self):
        mongouri = "mongodb://"
        if self.settings.get('username'):
            mongouri += self.settings.username
            if self.settings.get('password'):
                mongouri += ':' + self.settings.password
            mongouri += "@"
        mongouri += "%s:%s/%s" % (self.settings.host, self.settings.port,
                                  self.settings.database)

        return MongoClient(mongouri)

    def get(self, remote_path, local_path):
        client = self.connect()
        fs = GridFS(client[self.settings.database])
        try:
            fsfile = fs.get_last_version(filename=remote_path)
            with open(local_path, "wb") as dest:
                log.info("[GET] Copying mongodb:%s to %s" % (remote_path,
                                                             local_path))
                while True:
                    data = fsfile.read(4096)
                    if not data:
                        break
                    dest.write(data)
        except NoFile:
            client.disconnect()
            log.error("[GET] File not found: mongodb:%s" % remote_path)
            raise FileNotFound("%s not found" % remote_path)
        except Exception as e:
            log.error("[GET] Unhandled error: %s" % e)
            client.disconnect()
            raise StorageError(e)

    def put(self, local_path, remote_path):
        client = self.connect()
        fs = GridFS(client[self.settings.database])

        if self.exists(remote_path):
            raise FileAlreadyExists("%s already exists" % remote_path)

        gridin = fs.new_file(filename=remote_path)
        try:
            with open(local_path, "rb") as source:
                log.info("[PUT] Copying %s to mongodb:%s" % (local_path,
                                                             remote_path))
                while True:
                    data = source.read(4096)
                    if not data:
                        break
                    gridin.write(data)
                gridin.close()
            client.disconnect()
        except Exception as e:
            log.error("[PUT] Unhandled error: %s" % e)
            client.disconnect()
            raise StorageError(e)

    def delete(self, remote_path):
        client = self.connect()
        fs = GridFS(client[self.settings.database])
        try:
            fsfile = fs.get_last_version(filename=remote_path)
            fs.delete(fsfile._id)
            log.info("[DELETE] File deleted: mongodb:%s" % remote_path)
        except NoFile:
            client.disconnect()
            log.error("[DELETE] File not found: mongodb:%s" % remote_path)
            raise FileNotFound("%s not found" % remote_path)
        except Exception as e:
            log.error("[DELETE] Unhandled error: %s" % e)
            client.disconnect()
            raise StorageError(e)

    def exists(self, remote_path):
        client = self.connect()
        fs = GridFS(client[self.settings.database])
        ret = fs.exists(filename=remote_path)
        client.disconnect()
        return ret

    def size(self, remote_path):
        client = self.connect()
        fs = GridFS(client[self.settings.database])
        try:
            fsfile = fs.get_last_version(filename=remote_path)
        except NoFile as e:
            client.disconnect()
            log.error("[DELETE] File not found: mongodb:%s" % remote_path)
            raise FileNotFound(e)
        except Exception as e:
            log.error("[DELETE] Unhandled error: %s" % e)
            client.disconnect()
            raise StorageError(e)
        else:
            size = fsfile.length
            client.disconnect()
            return size

    def mtime(self, remote_path):
        client = self.connect()
        fs = GridFS(client[self.settings.database])
        try:
            fsfile = fs.get_last_version(filename=remote_path)
        except NoFile as e:
            client.disconnect()
            log.error("[DELETE] File not found: mongodb:%s" % remote_path)
            raise FileNotFound(e)
        except Exception as e:
            log.error("[DELETE] Unhandled error: %s" % e)
            client.disconnect()
            raise StorageError(e)
        else:
            timestamp = fsfile.upload_date
            client.disconnect()
            return timestamp



# -*- coding: utf-8 -*-
"""
    :copyright: Copyright 2013-2014 by Łukasz Mierzwa
    :contact: l.mierzwa@gmail.com
"""


import requests

from slumber import API


class UpaasAPI(API):

    def __init__(self, login, apikey, *args, **kwargs):
        session = requests.Session()
        session.headers.update({
            'X-UPAAS-LOGIN': login,
            'X-UPAAS-APIKEY': apikey,
        })
        kwargs['session'] = session
        super(UpaasAPI, self).__init__(*args, **kwargs)



# -*- coding: utf-8 -*-
"""
    :copyright: Copyright 2013-2014 by Łukasz Mierzwa
    :contact: l.mierzwa@gmail.com
"""


from __future__ import unicode_literals

import os
import time
import datetime

import pytest

from pymongo import MongoClient

from upaas.storage.mongodb import MongoDBStorage
from upaas.storage.exceptions import FileNotFound
from upaas.storage.utils import find_storage_handler
from upaas.config.base import ConfigurationError


@pytest.fixture(scope="module")
def storage(request):
    dbname = "testrun-%d" % time.time()
    storage = MongoDBStorage({"database": dbname})

    def cleanup():
        client = MongoClient()
        client.drop_database(dbname)
    request.addfinalizer(cleanup)

    return storage


def test_find_storage():
    assert find_storage_handler(
        'upaas.storage.mongodb.MongoDBStorage') is not None


def test_valid_settings(storage):
    assert storage.settings.host == "localhost"
    assert storage.settings.port == 27017


def test_invalid_settings():
    with pytest.raises(ConfigurationError):
        MongoDBStorage({"port": "string"})


def test_not_exists(storage):
    assert storage.exists("missing_file") is False


def test_get_not_exists(storage):
    with pytest.raises(FileNotFound):
        storage.get("missing file", "output")


def test_put_and_get(storage, empty_dir, empty_file):
    storage.put(empty_file, "abc")
    assert storage.exists("abc")

    local_path = os.path.join(empty_dir, "xyz")
    storage.get("abc", local_path)
    assert os.path.exists(local_path)


def test_delete_not_exists(storage):
    with pytest.raises(FileNotFound):
        storage.delete("missing file")


def test_put_and_delete(storage, empty_file):
    storage.put(empty_file, "delete.me")
    assert storage.exists("delete.me")

    storage.delete("delete.me")
    assert storage.exists("delete.me") is False


def test_size_empty(storage, empty_file):
    storage.put(empty_file, "empty.file")
    assert storage.size("empty.file") == 0


def test_size_not_empty(storage, empty_file):
    with open(empty_file, 'w') as f:
        f.write('123456789')

    storage.put(empty_file, "not-empty.file")
    assert storage.size("not-empty.file") == 9


def test_mtime(storage, empty_file):
    storage.put(empty_file, "mtime.check")
    assert isinstance(storage.mtime("mtime.check"), datetime.datetime)
    # FIXME basic check to verify if timestamp is from the past
    assert storage.mtime("mtime.check") < datetime.datetime.now()



# -*- coding: utf-8 -*-
"""
    :copyright: Copyright 2013-2014 by Łukasz Mierzwa
    :contact: l.mierzwa@gmail.com
"""


# code from
# http://www.thehosthelpers.com/how-tos-tips-tricks/protip-secure-your-rooted
# - python-script-(or-easy-chroot-class)/


from __future__ import unicode_literals

import os
import logging


log = logging.getLogger(__name__)


class Chroot(object):

    def __init__(self, root, workdir=None, umask=0o22):
        self.root = root
        self.dir = '/' if workdir is None else workdir
        self.umask = umask

    def __enter__(self):
        log.debug("Entering chroot at '%s', workdir is '%s'" % (self.root,
                                                                self.dir))
        self.realdir = os.getcwd()
        self.realroot = os.open('/', os.O_RDONLY)
        os.chroot(self.root)
        os.chdir(self.dir)
        self.old_umask = os.umask(self.umask)
        return self

    def __exit__(self, type, value, traceback):
        try:
            os.fchdir(self.realroot)
        except OSError:
            return
        os.chroot('.')
        os.close(self.realroot)
        os.chdir(self.realdir)
        os.umask(self.old_umask)
        log.debug("Exited from chroot at '%s'" % self.root)

    def escape(self):
        try:
            os.fchdir(self.realroot)
        except OSError:
            return
        while os.stat('.')[1] != os.stat('..')[1]:
            os.chdir('..')
        os.chroot('.')
        os.close(self.realroot)
        os.chdir(self.realdir)
        os.umask(self.old_umask)



# -*- coding: utf-8 -*-
"""
    :copyright: Copyright 2013-2014 by Łukasz Mierzwa
    :contact: l.mierzwa@gmail.com
"""


from __future__ import unicode_literals

import sys
import codecs
import logging
try:
    from urllib.parse import urljoin
except ImportError:
    from urlparse import urljoin

from plumbum import cli

from slumber.exceptions import SlumberHttpBaseException

from upaas.client import UpaasAPI


class UPaaSApplication(cli.Application):

    api = None
    log = None
    log_level = "warning"
    log_output = "-"

    @cli.switch(["l", "log-level"], str, help="Logging level")
    def set_log_level(self, level):
        self.log_level = level

    @cli.switch(["o", "log-output"], str, help="Log file ('-' for console)")
    def set_log_output(self, output):
        self.log_output = output

    def setup_logger(self):
        if self.log_output == "-":
            handler = logging.StreamHandler(
                codecs.getwriter('utf-8')(sys.stdout))
        else:
            handler = logging.FileHandler(self.log_output, encoding='utf-8')

        handler.setFormatter(
            logging.Formatter("%(asctime)s %(levelname)s: %(message)s",
                              "%Y-%m-%d %H:%M:%S"))
        self.log = logging.getLogger()
        if not len(self.log.handlers):
            self.log.addHandler(handler)

            self.log.setLevel(getattr(logging, self.log_level.upper(),
                                      logging.WARNING))

    def api_connect(self, login, apikey, url):
        if not self.api:
            url = urljoin(url, '/api/v1/')
            self.log.info("Connecting to API at '%s' as '%s'" % (url, login))
            self.api = UpaasAPI(login, apikey, url)

    def print_msg(self, msg, prefix=">>"):
        if prefix:
            prefix += " "
        for line in msg.splitlines():
            print(("%s%s" % (prefix or '', line)))

    def handle_error(self, err):
        if isinstance(err, SlumberHttpBaseException):
            if err.response.status_code == 401:
                self.log.error("Authentication error")
            elif err.content:
                self.log.error(err.content)
            else:
                self.log.error(err)
        else:
            self.log.error(err)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
    :copyright: Copyright 2013-2014 by Łukasz Mierzwa
    :contact: l.mierzwa@gmail.com
"""


from __future__ import unicode_literals

from setuptools import setup, find_packages

try:
    from pip.req import parse_requirements
    required = {'install_requires': [str(r.req) for r in parse_requirements(
        'requirements.txt')]}
except ImportError:
    required = {}


setup(
    name='upaas-common',
    version='0.3.1',
    license='GPLv3',
    description='uPaaS common classes',
    author='Łukasz Mierzwa',
    author_email='l.mierzwa@gmail.com',
    url='https://github.com/prymitive/upaas-common',
    packages=find_packages(exclude=["tests"]),
    classifiers=[
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Programming Language :: Python',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    platforms=['Linux'],
    **required
)



# -*- coding: utf-8 -*-
"""
    :copyright: Copyright 2013-2014 by Łukasz Mierzwa
    :contact: l.mierzwa@gmail.com
"""


from upaas.config import base


class BaseStorage(object):

    configuration_schema = {}

    def __init__(self, settings):
        self.settings = base.Config(settings,
                                    _schema=self.configuration_schema)

    def get(self, remote_path, local_path):
        """
        Get file from storage.

        :param remote_path: Path of the file we want to download from storage.
        :param local_path: Where to put downloaded file.
        """
        raise NotImplementedError

    def put(self, local_path, remote_path):
        """
        Upload file to storage.

        :param local_path: Path of the file to upload.
        :param remote_path: Path under uploaded file should be available.
        """
        raise NotImplementedError

    def delete(self, remote_path):
        """
        Delete file from storage.

        :param remote_path: Path of the remote file to be deleted from storage.
        """
        raise NotImplementedError

    def exists(self, remote_path):
        """
        Check if given path (file or directory) exists on storage.

        :param remote_path: Path of the remote file to be checked.
        """
        raise NotImplementedError

    def size(self, remote_path):
        """
        Return size of the file.

        :param remote_path: Path of the remote file.
        """
        raise NotImplementedError

    def mtime(self, remote_path):
        """
        Return modification time of the file as datetime object.

        :param remote_path: Path of the remote file to be checked.
        """
        raise NotImplementedError



# -*- coding: utf-8 -*-
"""
    :copyright: Copyright 2013-2014 by Łukasz Mierzwa
    :contact: l.mierzwa@gmail.com
"""


from netifaces import interfaces, ifaddresses, AF_INET


def local_ipv4_addresses():
    ip_list = []
    for interface in interfaces():
        for link in ifaddresses(interface).get(AF_INET, []):
            if link['addr'] != '127.0.0.1':
                ip_list.append(link['addr'])
    return ip_list
